''' Glyph Positioning System (GPOS) tables '''

from __future__ import annotations
from typing import Optional
from collections import namedtuple

from .fontread import FontReader


class Gpos:
    ''' Glyph Positioning System Table '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        self.fontfile.seek(self.ofst)

        self.vermajor = self.fontfile.readuint16()
        self.verminor = self.fontfile.readuint16()
        scriptofst = self.fontfile.readuint16()
        featureofst = self.fontfile.readuint16()
        lookupofst = self.fontfile.readuint16()
        if self.verminor > 0:
            self.variationofst = self.fontfile.readuint32()

        # Read scripts
        scriptlisttableloc = self.ofst + scriptofst
        scriptcnt = self.fontfile.readuint16(scriptlisttableloc)
        self.scripts = {}
        for i in range(scriptcnt):
            tag = self.fontfile.read(4).decode()
            self.scripts[tag] = Script(
                tag,
                self.fontfile.readuint16() + scriptlisttableloc,
                self.fontfile)

        # Read features
        featurelisttableloc = self.ofst + featureofst
        featurecnt = self.fontfile.readuint16(featurelisttableloc)
        self.features = []
        for i in range(featurecnt):
            self.features.append(Feature(
                self.fontfile.read(4).decode(),
                self.fontfile.readuint16() + featurelisttableloc,
                self.fontfile))

        # Read Lookups
        lookuplisttableloc = self.ofst + lookupofst
        lookupcnt = self.fontfile.readuint16(lookuplisttableloc)
        self.lookups = []
        for i in range(lookupcnt):
            self.lookups.append(Lookup(
                self.fontfile.readuint16() + lookuplisttableloc,
                self.fontfile))

    def kern(self, glyph1: int, glyph2: int, script:
             str=None, lang: str=None) -> tuple[Optional[dict], Optional[dict]]:
        ''' Get kerning adjustmnet for glyph1 and glyph2 '''
        scr = self.scripts.get(script, self.scripts.get('DFLT', self.scripts.get('latn')))  # type: ignore
        if scr is None:
            return {}, {}
        langsys = scr.languages.get(lang, scr.languages.get('DFLT'))  # type: ignore
        if langsys is None:
            return {}, {}

        # Find kerning features in features list
        usefeatures = [self.features[i] for i in langsys.featureidxs]  # type: ignore
        featnames = [f.tag for f in usefeatures]
        if 'kern' in featnames:
            lookups = usefeatures[featnames.index('kern')].lookupids
            tables = [self.lookups[i] for i in lookups]
            for table in tables:
                for subtable in table.subtables:
                    v1, v2 = subtable.get_adjust(glyph1, glyph2)
                    if v1 or v2:
                        return v1, v2
        return {}, {}  # No adjustments

    def __repr__(self):
        return f'<GPOS Table v{self.vermajor}.{self.verminor}>'


class Script:
    ''' GPOS Script table '''
    def __init__(self, tag: str, ofst: int, fontfile: FontReader):
        self.tag = tag
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()
        self.fontfile.seek(self.ofst)
        self.defaultLangSysOfst = self.fontfile.readuint16()
        self.languages = {}

        if self.defaultLangSysOfst:
            self.languages['DFLT'] = LanguageSystem(
                'DFLT',
                self.defaultLangSysOfst + self.ofst,
                self.fontfile)

        langsyscnt = self.fontfile.readuint16()
        for i in range(langsyscnt):
            tag = self.fontfile.read(4).decode()
            self.languages[tag] = (LanguageSystem(
                tag,
                self.fontfile.readuint16() + self.ofst,
                self.fontfile))

        self.fontfile.seek(fileptr)  # Put file pointer back

    def __repr__(self):
        return f'<Script {self.tag}, {hex(self.ofst)}>'


class LanguageSystem:
    ''' GPOS Language System Table '''
    def __init__(self, tag: str, ofst: int, fontfile: FontReader):
        self.tag = tag
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()
        self.fontfile.seek(self.ofst)
        self.fontfile.readuint16()  # lookupordeoffset (reserved)
        self.reqdFeatureIndex = self.fontfile.readuint16()
        featurecount = self.fontfile.readuint16()
        self.featureidxs = []
        for i in range(featurecount):
            self.featureidxs.append(self.fontfile.readuint16())

        self.fontfile.seek(fileptr)  # Put file pointer back

    def __repr__(self):
        return f'<LanguageSystem {self.tag}, {hex(self.ofst)}>'


class Feature:
    ''' GPOS Feature Table '''
    def __init__(self, tag: str, ofst: int, fontfile: FontReader):
        self.tag = tag
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()
        self.fontfile.seek(self.ofst)
        self.featureparamofst = self.fontfile.readuint16()
        lookupcnt = self.fontfile.readuint16()
        self.lookupids = []
        for i in range(lookupcnt):
            self.lookupids.append(self.fontfile.readuint16())

        self.fontfile.seek(fileptr)  # Put file pointer back

    def __repr__(self):
        return f'<Feature {self.tag}, {hex(self.ofst)}>'


class Lookup:
    ''' GPOS Lookup Table '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()

        RIGHT_TO_LEFT = 0x0001
        IGNORE_BASE_GLYPHS = 0x0002
        IGNORE_LIGATURES = 0x0004
        IGNORE_MARKS = 0x0008
        USE_MARK_FILTERING_SET = 0x0010
        MARK_ATTACHMENT_TYPE_MASK = 0xFF00

        self.fontfile.seek(self.ofst)
        self.type = self.fontfile.readuint16()
        self.flag = self.fontfile.readuint16()
        subtablecnt = self.fontfile.readuint16()
        self.tableofsts = []
        for i in range(subtablecnt):
            self.tableofsts.append(self.fontfile.readuint16())
        self.markfilterset = None
        if self.flag & USE_MARK_FILTERING_SET:
            self.markfilterset = self.fontfile.readuint16()

        self.subtables = []
        if self.type == 2:  # Pair adjustment positioning
            for i in range(subtablecnt):
                self.subtables.append(PairAdjustmentTable(
                    self.tableofsts[i] + self.ofst,
                    self.fontfile))
        elif self.type == 9:  # Extension subtable. Can be any type
            ptr = self.fontfile.tell()
            posformat = self.fontfile.readuint16()
            assert posformat == 1
            exttype = self.fontfile.readuint16()
            extofst = self.fontfile.readuint32()
            if exttype == 2:
                self.subtables.append(PairAdjustmentTable(
                    ptr + extofst,
                    self.fontfile))
        else:
            pass ## print('Lookup', self.type)

        self.fontfile.seek(fileptr)  # Put file pointer back

    def __repr__(self):
        return f'<LookupTable {hex(self.ofst)}>'


class PairAdjustmentTable:
    ''' Pair Adjustment Table - informs kerning between pairs of glyphs '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()

        self.fontfile.seek(self.ofst)
        self.posformat = self.fontfile.readuint16()
        self.covofst = self.fontfile.readuint16()
        self.valueformat1 = self.fontfile.readuint16()
        self.valueformat2 = self.fontfile.readuint16()

        if self.posformat == 1:
            pairsetcount = self.fontfile.readuint16()
            pairsetofsts = []
            for i in range(pairsetcount):
                pairsetofsts.append(self.fontfile.readuint16())

            PairValue = namedtuple('PairValue', ['secondglyph', 'value1', 'value2'])
            self.pairsets = []
            for i in range(pairsetcount):
                self.fontfile.seek(pairsetofsts[i] + self.ofst)
                paircnt = self.fontfile.readuint16()
                pairs = []
                for p in range(paircnt):
                    pairs.append(PairValue(
                        self.fontfile.readuint16(),
                        self.fontfile.readvaluerecord(self.valueformat1),
                        self.fontfile.readvaluerecord(self.valueformat2)))
                self.pairsets.append(pairs)

        elif self.posformat == 2:
            classdef1ofst = self.fontfile.readuint16()
            classdef2ofst = self.fontfile.readuint16()
            class1cnt = self.fontfile.readuint16()
            class2cnt = self.fontfile.readuint16()

            self.classrecords = []
            for i in range(class1cnt):
                class2recs = []
                for j in range(class2cnt):
                    class2recs.append(
                        (self.fontfile.readvaluerecord(self.valueformat1),
                         self.fontfile.readvaluerecord(self.valueformat2)))
                self.classrecords.append(class2recs)

            self.class1def = ClassDef(self.ofst + classdef1ofst, self.fontfile)
            self.class2def = ClassDef(self.ofst + classdef2ofst, self.fontfile)

        else:
            raise ValueError('Invalid posformat in PairAdjustmnetTable')

        self.coverage = Coverage(self.covofst+self.ofst, self.fontfile)
        self.fontfile.seek(fileptr)  # Put file pointer back

    def get_adjust(self, glyph1: int, glyph2: int) -> tuple[Optional[dict], Optional[dict]]:
        ''' Get kerning adjustment for glyph1 and glyph2 pair '''
        v1 = v2 = None

        # Look up first glyph in coverage table
        covidx = self.coverage.covidx(glyph1)
        if covidx is not None:

            # Look up second glyph
            if self.posformat == 1:
                for p in self.pairsets[covidx]:
                    if p.secondglyph == glyph2:
                        v1 = p.value1
                        v2 = p.value2
                        break

            else:
                c1 = self.class1def.get_class(glyph1)
                c2 = self.class2def.get_class(glyph2)
                if c1 is not None and c2 is not None:
                    v1, v2 = self.classrecords[c1][c2]

        return v1, v2

    def __repr__(self):
        return f'<PairAdjustmentTable {hex(self.ofst)}>'


class Coverage:
    ''' Coverage Table - defines which glyphs apply to this lookup '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()

        self.fontfile.seek(self.ofst)
        self.format = self.fontfile.readuint16()

        if self.format == 1:
            glyphcnt = self.fontfile.readuint16()
            self.glyphs = []
            for i in range(glyphcnt):
                self.glyphs.append(self.fontfile.readuint16())

        elif self.format == 2:
            rangecnt = self.fontfile.readuint16()
            Range = namedtuple('Range', ['startglyph', 'endglyph', 'covidx'])
            self.ranges = []
            for i in range(rangecnt):
                self.ranges.append(Range(
                    self.fontfile.readuint16(),
                    self.fontfile.readuint16(),
                    self.fontfile.readuint16()))
        else:
            raise ValueError('Bad coverage table format')

        self.fontfile.seek(fileptr)  # Put file pointer back

    def covidx(self, glyph: int) -> Optional[int]:
        ''' Get coverage index for this glyph, or None if not in the coverage range '''
        if self.format == 1:
            try:
                idx: Optional[int] = self.glyphs.index(glyph)
            except (ValueError, IndexError):
                idx = None
        else:
            for r in self.ranges:
                if r.startglyph <= glyph <= r.endglyph:
                    idx = r.covidx + (glyph - r.startglyph)
                    break
            else:
                idx = None
        return idx

    def __repr__(self):
        return f'<CoverageTable {hex(self.ofst)}>'


class ClassDef:
    ''' Class Definition Table - defines a "class" of glyphs. '''
    def __init__(self, ofst: int, fontfile: FontReader):
        self.ofst = ofst
        self.fontfile = fontfile
        fileptr = self.fontfile.tell()

        self.fontfile.seek(self.ofst)
        self.format = self.fontfile.readuint16()

        if self.format == 1:
            self.startglyph = self.fontfile.readuint16()
            glyphcnt = self.fontfile.readuint16()
            self.classvalues = []
            for i in range(glyphcnt):
                self.classvalues.append(self.fontfile.readuint16())

        elif self.format == 2:
            ClassRange = namedtuple('ClassRange', ['startglyph', 'endglyph', 'cls'])
            rangecnt = self.fontfile.readuint16()
            self.ranges = []
            for i in range(rangecnt):
                self.ranges.append(ClassRange(
                    self.fontfile.readuint16(),
                    self.fontfile.readuint16(),
                    self.fontfile.readuint16()))
        else:
            raise ValueError

        self.fontfile.seek(fileptr)  # Put file pointer back

    def get_class(self, glyphid: int) -> int:
        ''' Get class by glyph id '''
        if self.format == 1:
            if self.startglyph <= glyphid < self.startglyph + len(self.classvalues):
                return self.classvalues[glyphid - self.startglyph]
            else:
                return 0

        else:
            for i, rng in enumerate(self.ranges):
                if rng.startglyph <= glyphid <= rng.endglyph:
                    return rng.cls
            return 0

    def __repr__(self):
        return f'<ClassDefTable {hex(self.ofst)}>'



''' CLass for reading bytes from a font file '''

import struct
from datetime import datetime, timedelta
from io import BytesIO


class FontReader(BytesIO):
    ''' Class for reading from Font File '''
    # TTF/OTF is big-endian (>)

    def readuint32(self, offset: int=None) -> int:
        ''' Read 32-bit unsigned integer '''
        if offset:
            self.seek(offset)
        return struct.unpack('>I', self.read(4))[0]

    def readuint16(self, offset: int=None) -> int:
        ''' Read 16-bit unsigned integer '''
        if offset:
            self.seek(offset)
        return struct.unpack('>H', self.read(2))[0]

    def readuint8(self, offset: int=None) -> int:
        ''' Read 8-bit unsigned integer '''
        if offset:
            self.seek(offset)
        return struct.unpack('>B', self.read(1))[0]

    def readint8(self, offset: int=None) -> int:
        ''' Read 8-bit signed integer '''
        if offset:
            self.seek(offset)
        return struct.unpack('>b', self.read(1))[0]

    def readdate(self, offset: int=None) -> datetime:
        ''' Read datetime '''
        if offset:
            self.seek(offset)
        mtime = self.readuint32() * 0x100000000 + self.readuint32()
        fontepoch = datetime(1904, 1, 1)
        mdate = fontepoch + timedelta(seconds=mtime)
        return mdate

    def readint16(self, offset: int=None) -> int:
        ''' Read 16-bit signed integer '''
        if offset:
            self.seek(offset)
        return struct.unpack('>h', self.read(2))[0]

    def readshort(self, offset: int=None) -> float:
        ''' Read "short" fixed point number (S1.14) '''
        x = self.readint16()
        return float(x) * 2**-14

    def readvaluerecord(self, fmt: int) -> dict:
        ''' Read a GPOS "ValueRecord" into a dictionary. Zero values will be omitted. '''
        record = {}
        if fmt & 0x0001:
            record['x'] = self.readint16()
        if fmt & 0x0002:
            record['y'] = self.readint16()
        if fmt & 0x0004:
            record['xadvance'] = self.readint16()
        if fmt & 0x0008:
            record['yadvance'] = self.readint16()
        if fmt & 0x0010:
            record['xpladeviceoffset'] = self.readuint16()
        if fmt & 0x0020:
            record['ypladeviceoffset'] = self.readuint16()
        if fmt & 0x0040:
            record['xadvdeviceoffset'] = self.readuint16()
        if fmt & 0x0080:
            record['yadvdeviceoffset'] = self.readuint16()
        return record



''' Read font file and write glyphs to SVG '''

from __future__ import annotations
from typing import Literal, Union, Optional, Dict
import sys
import os
import math
from pathlib import Path
from collections import namedtuple
import importlib.resources as pkg_resources
import xml.etree.ElementTree as ET
import warnings

from .config import config
from .fontread import FontReader
from . import gpos
from .cmap import Cmap12, Cmap4
from .glyph import read_glyph, fmt, SimpleGlyph, CompoundGlyph
from .fonttypes import AdvanceWidth, Layout, Header, Table, FontInfo, FontNames, Symbols


fontindex = None


class Font:
    ''' Class to read/parse a OpenType/TTF and write glyphs to SVG

        Args:
            name: File name of the font
            style: Font style such as "bold" or "italic", used when searching
                system paths for a font file
    '''
    def __init__(self, name: Union[str, Path]=None, style: str='regular'):
        self.fname = None
        if name and Path(name).exists():
            self.fname = Path(name)
        elif name:
            self.fname = findfont(name, style)
            if self.fname is None:
                warnings.warn(f'Font {name} not found.')

        if self.fname is None:
            with pkg_resources.path('ziafont.fonts', 'DejaVuSans.ttf') as p:
                self.fname = p

        with open(self.fname, 'rb') as f:
            self.fontfile = FontReader(f.read())
        self.info = self._loadfont()  # Load in all the font metadata
        self._glyphs: Dict[int, Union[SimpleGlyph, CompoundGlyph]] = {}
        self._glyphids: Dict[str, int] = {}

    def _loadfont(self) -> FontInfo:
        ''' Read font metadata '''
        self._readtables()
        header, layout = self._readheader()
        self.advwidths = self._readwidths(header.numlonghormetrics)
        self._readcmap()
        names = self._readnames()
        info = FontInfo(self.fname, names, header, layout)

        self.gpos = None
        if 'GPOS' in self.tables:
            self.gpos = gpos.Gpos(self.tables['GPOS'].offset, self.fontfile)

        return info

    def _readtables(self) -> None:
        ''' Read list of tables in the font, and verify checksums '''
        self.fontfile.seek(0)
        scalartype = self.fontfile.readuint32()
        numtables = self.fontfile.readuint16()
        searchrange = self.fontfile.readuint16()
        entryselector = self.fontfile.readuint16()
        rangeshift = self.fontfile.readuint16()   # numtables*16-searchrange

        # Table Directory (table 5)
        self.tables = {}
        for i in range(numtables):
            tag = self.fontfile.read(4).decode()
            self.tables[tag] = Table(checksum=self.fontfile.readuint32(),
                                     offset=self.fontfile.readuint32(),
                                     length=self.fontfile.readuint32())

        if 'glyf' not in self.tables:
            raise ValueError('Unsupported font (no glyf table).')

    def verifychecksum(self):
        ''' Verify checksum of all font tables. Raises ValueError if invalid '''
        for table in self.tables.keys():
            if table != 'head':
                self._verifychecksum(table)

    def _verifychecksum(self, table: str) -> None:
        ''' Verify checksum of table. Raises ValueError if invalid. '''
        tb = self.tables[table]
        self.fontfile.seek(tb.offset)
        s = 0
        nlongs = (tb.length + 3) // 4
        s = sum(self.fontfile.readuint32() for i in range(nlongs)) & 0xffffffff
        if s != tb.checksum:
            raise ValueError(f'Table {table} checksum {s} != saved checksum {tb.checksum}')

    def _readheader(self) -> tuple[Header, Layout]:
        ''' Read Font "head" and "hhea" tables '''
        version = self.fontfile.readuint32(self.tables['head'].offset)
        revision = self.fontfile.readuint32()
        chksumadjust = self.fontfile.readuint32()
        magic = self.fontfile.readuint32()
        assert magic == 0x5f0f3cf5
        flags = self.fontfile.readuint16()
        unitsperem = self.fontfile.readuint16()
        created = self.fontfile.readdate()
        modified = self.fontfile.readdate()
        xmin = self.fontfile.readint16()
        ymin = self.fontfile.readint16()
        xmax = self.fontfile.readint16()
        ymax = self.fontfile.readint16()
        macstyle = self.fontfile.readuint16()
        lowestrecppem = self.fontfile.readuint16()
        directionhint = self.fontfile.readint16()
        indextolocformat = self.fontfile.readint16()
        glyphdataformat = self.fontfile.readint16()

        # hhea table with other parameters
        _ = self.fontfile.readuint32(self.tables['hhea'].offset)  # version
        ascent = self.fontfile.readint16()
        descent = self.fontfile.readint16()
        linegap = self.fontfile.readint16()
        advwidthmax = self.fontfile.readuint16()
        minleftbearing = self.fontfile.readint16()
        minrightbearing = self.fontfile.readint16()
        xmaxextent = self.fontfile.readint16()
        caretsloperise = self.fontfile.readint16()
        caretsloperun = self.fontfile.readint16()
        caretoffset = self.fontfile.readint16()
        for i in range(4):
            self.fontfile.readint16()  # Skip reserved
        metricformat = self.fontfile.readint16()
        numlonghormetrics = self.fontfile.readuint16()

        advwidth = AdvanceWidth(advwidthmax, minleftbearing)

        layout = Layout(unitsperem, xmin, xmax, ymin, ymax, ascent, descent,
                        advwidth, minleftbearing, minrightbearing)
        header = Header(version, revision, chksumadjust, magic, flags,
                        created, modified, macstyle,
                        lowestrecppem, directionhint, indextolocformat,
                        glyphdataformat, numlonghormetrics)
        return header, layout

    def _readnames(self) -> FontNames:
        ''' Read the "name" table which includes the font name,
            copyright, and other info
        '''
        namefmt = self.fontfile.readuint16(self.tables['name'].offset)
        nameids = [''] * 15  # Empty strings for nameId table

        if namefmt == 0:  # '1' is not supported
            count = self.fontfile.readuint16()
            strofst = self.fontfile.readuint16()
            namerecords = []
            for i in range(count):
                platformId = self.fontfile.readuint16()
                platformSpecificId = self.fontfile.readuint16()
                languageId = self.fontfile.readuint16()
                nameId = self.fontfile.readuint16()
                length = self.fontfile.readuint16()
                offset = self.fontfile.readuint16()
                namerecords.append((platformId, platformSpecificId, languageId,
                                    nameId, length, offset))
            for record in namerecords:
                self.fontfile.seek(self.tables['name'].offset + strofst + record[5])
                name = self.fontfile.read(record[4])
                if record[3] < 16:
                    if record[0] in [0, 3]:  # Microsoft and Unicode formats
                        nameids[record[3]] = name.decode('utf-16be')

        return FontNames(*nameids)

    def _readwidths(self, numlonghormetrics: int) -> list[AdvanceWidth]:
        ''' Read `advanceWidth` and `leftsidebearing` from "htmx" table '''
        self.fontfile.seek(self.tables['hmtx'].offset)
        advwidths = []
        for i in range(numlonghormetrics):
            w = self.fontfile.readuint16()
            b = self.fontfile.readint16()
            advwidths.append(AdvanceWidth(w, b))
        return advwidths

    def _readcmap(self) -> None:
        ''' Read "cmap" table and select a cmap for locating glyphs from characters.
            Cmap formats 4 and 12 are supported.
        '''
        platforms = {0: 'unicode', 1: 'macintosh', 3: 'windows'}
        version = self.fontfile.readint16(self.tables['cmap'].offset)
        numtables = self.fontfile.readint16()
        CMapTable = namedtuple('CMapTable', ['platform', 'platformid', 'offset'])
        cmaptables = []
        for i in range(numtables):
            cmaptables.append(CMapTable(
                platforms.get(self.fontfile.readuint16()),
                self.fontfile.readuint16(),
                self.fontfile.readuint32()))

        self.cmap: Optional[Union[Cmap12, Cmap4]] = None  # Active cmap
        self.cmaps: list[Union[Cmap12, Cmap4]] = []
        cmap: Union[Cmap12, Cmap4]
        for ctable in cmaptables:
            cmapformat = self.fontfile.readuint16(self.tables['cmap'].offset + ctable.offset)
            if cmapformat == 4:
                endcodes = []
                startcodes = []
                iddeltas = []
                idrangeoffset = []
                glyphidxarray = []
                length = self.fontfile.readuint16()
                lang = self.fontfile.readuint16()
                segcount = self.fontfile.readuint16() // 2
                searchrange = self.fontfile.readuint16()
                entryselector = self.fontfile.readuint16()
                rangeshift = self.fontfile.readuint16()
                for i in range(segcount):
                    endcodes.append(self.fontfile.readuint16())
                _ = self.fontfile.readuint16()  # reserved pad
                for i in range(segcount):
                    startcodes.append(self.fontfile.readuint16())
                for i in range(segcount):
                    iddeltas.append(self.fontfile.readuint16())
                for i in range(segcount):
                    idrangeoffset.append(self.fontfile.readuint16())

                # Length of glyph array comes from total length of cmap table
                # //2 because len is in bytes, but table glyphidxarray is 16-bit
                glyphtablelen = (length - (self.fontfile.tell() - self.tables['cmap'].offset)) // 2
                for i in range(glyphtablelen):
                    glyphidxarray.append(self.fontfile.readuint16())
                cmap = Cmap4(ctable.platform, ctable.platformid,
                             startcodes, endcodes, idrangeoffset, iddeltas, glyphidxarray)
                if self.cmap is None:
                    self.cmap = cmap
                self.cmaps.append(cmap)

            elif cmapformat == 12:
                _ = self.fontfile.readuint16()
                length = self.fontfile.readuint32()
                lang = self.fontfile.readuint32()
                ngroups = self.fontfile.readuint32()
                starts = []
                ends = []
                glyphstarts = []
                for i in range(ngroups):
                    starts.append(self.fontfile.readuint32())
                    ends.append(self.fontfile.readuint32())
                    glyphstarts.append(self.fontfile.readuint32())
                cmap = Cmap12(ctable.platform, ctable.platformid, starts, ends, glyphstarts)
                if self.cmap is None or isinstance(self.cmap, Cmap4):
                    self.cmap = cmap
                self.cmaps.append(cmap)

        if len(self.cmaps) == 0:
            raise ValueError('No suitable cmap table found in font.')

    def usecmap(self, cmapidx: int) -> None:
        ''' Select cmap table by index. Only supported tables are included. '''
        self.cmap = self.cmaps[cmapidx]

    def glyphindex(self, char: str) -> int:
        ''' Get index of character glyph '''
        gid = self._glyphids.get(char)
        if gid is None:
            gid = self.cmap.glyphid(char)  # type: ignore
            self._glyphids[char] = gid
        return gid

    def _glyphoffset(self, index: int) -> Optional[int]:
        ''' Get offset (from beginning of file) of glyph,
            Return None if no glyph (empty/space) at this index.
        '''
        if self.info.header.indextolocformat == 1:
            offset = self.fontfile.readuint32(self.tables['loca'].offset + index * 4)
            nextofst = self.fontfile.readuint32()
        else:
            offset = self.fontfile.readuint16(self.tables['loca'].offset + index * 2) * 2
            nextofst = self.fontfile.readuint16() * 2

        if offset == nextofst:
            # Empty glyphs (ie space) have no length.
            return None
        else:
            return offset + self.tables['glyf'].offset

    def glyph(self, char: str) -> SimpleGlyph:
        ''' Get the Glyph for the character '''
        index = self.glyphindex(char)        # Glyph Number
        return self.glyph_fromid(index)

    def glyph_fromid(self, glyphid: int) -> Union[SimpleGlyph, CompoundGlyph]:
        ''' Read a glyph from the "glyf" table

            Args:
                glyphid: Glyph index used to find glyph data
        '''
        glyph = self._glyphs.get(glyphid)
        if glyph is None:
            glyph = read_glyph(glyphid, self)
            self._glyphs[glyphid] = glyph
        return glyph

    def advance(self, glyph1: int, glyph2: int=None, kern: bool=True):
        ''' Get advance width in font units, including kerning adjustment if glyph2 is defined '''
        try:
            adv = self.advwidths[glyph1].width
        except IndexError:
            adv = self.info.layout.advwidthmax.width

        if kern and glyph2 and self.gpos:
            # Only getting x-advance for first glyph.
            adv += self.gpos.kern(glyph1, glyph2)[0].get('xadvance', 0)
        return adv

    def getsize(self, s) -> tuple[float, float]:
        ''' Calculate width and height (including ascent/descent) of string '''
        txt = Text(s, self)
        return txt.getsize()

    def text(self, s: str,
             size: float=None,
             linespacing: float=1,
             halign: Literal['left', 'center', 'right']='left',
             valign: Literal['base', 'center', 'top']='base',
             color: str=None,
             rotation: float=0,
             rotation_mode: str='anchor',
             kern=True):
        ''' Create a Text object using this font

            Args:
                s: String to convert.
                size: Font size in points
                linespacing: Space between lines
                halign: Horizontal Alignment
                valign: Vertical Alignment
                color: Color for string
                rotation: Rotation angle in degrees
                rotation_mode: Either 'default' or 'anchor', to
                    mimic Matplotlib behavoir. See:
                    https://matplotlib.org/stable/gallery/text_labels_and_annotations/demo_text_rotation_mode.html
                kern: Use font kerning adjustment
        '''
        txt = Text(s, self, size=size, linespacing=linespacing, halign=halign, valign=valign,
                   color=color, kern=kern, rotation=rotation,
                   rotation_mode=rotation_mode)
        return txt


class Text:
    ''' Draw text as SVG paths

        Args:
            s: String to draw
            font: Font name or ziafont.Font to use
            size: Font size in points
            linespacing: Spacing between lines
            color: Color for string
            halign: Horizontal Alignment
            valign: Vertical Alignment
            rotation: Rotation angle in degrees
            rotation_mode: Either 'default' or 'anchor', to
                mimic Matplotlib behavoir. See:
                https://matplotlib.org/stable/gallery/text_labels_and_annotations/demo_text_rotation_mode.html
            kern: Use kerning adjustment
    '''
    def __init__(self, s: str,  font: Union[str, Font]=None,
                 size: float=None, linespacing: float=1,
                 halign: Literal['left', 'center', 'right']='left',
                 valign: Literal['base', 'center', 'top']='base',
                 color: str=None,
                 rotation: float=0,
                 rotation_mode: str='anchor',
                 kern: bool=True):
        self.str = s
        self.halign = halign
        self.valign = valign
        self.color = color
        self.rotation = rotation
        self.rotation_mode = rotation_mode
        self.size = size if size else config.fontsize
        self.linespacing = linespacing
        self.kern = kern
        if font is None or isinstance(font, str):
            self.font = Font(font)
        else:
            self.font = font
        self._symbols = self._buildstring()

    def svgxml(self) -> ET.Element:
        ''' Get SVG XML element '''
        svg = ET.Element('svg')
        svg.attrib['xmlns'] = 'http://www.w3.org/2000/svg'
        if config.svg2:
            svg.attrib['xmlns:xlink'] = 'http://www.w3.org/1999/xlink'
        ret, (xmin, xmax, ymin, ymax) = self._drawon(svg)
        w = xmax-xmin
        h = ymax-ymin
        svg.attrib['width'] = fmt(w)
        svg.attrib['height'] = fmt(h)
        svg.attrib['viewBox'] = f'{fmt(xmin)} {fmt(ymin)} {fmt(w)} {fmt(h)}'
        return ret

    def svg(self) -> str:
        ''' Get SVG string '''
        return ET.tostring(self.svgxml(), encoding='unicode')

    def _repr_svg_(self):
        ''' Jupyter representer '''
        return self.svg()

    def drawon(self, svg: ET.Element, x: float=0, y: float=0):
        ''' Draw text on the SVG '''
        svg, _ = self._drawon(svg, x, y)
        return svg

    def _drawon(self, svg: ET.Element, x: float=0, y: float=0):
        ''' Draw text on the SVG '''
        word, symbols, width, ymin, ymax = self._symbols
        height = ymax-ymin
        # Adjust vertical alignment
        yofst = {'base': 0,
                 'bottom': -ymax,
                 'top': -ymin,
                 'center': -height/2-ymin}.get(self.valign, 0)
        xofst = {'center': -width/2,
                 'right': -width}.get(self.halign, 0)
        xy = x + xofst, y + yofst

        # Get existing symbol/glyphs, add ones not there yet
        if config.svg2:
            existingsymbols = svg.findall('symbol')
            symids = [sym.attrib.get('id') for sym in existingsymbols]
            for sym in symbols:
                if sym not in symids:
                    svg.append(sym)

        xform = ''
        if xy != (0, 0):
            xform = f'translate({fmt(xy[0])} {fmt(xy[1])})'

        svg.append(word)

        if config.debug:  # Test viewbox
            rect = ET.SubElement(word, 'rect')
            rect.attrib['x'] = '0'
            rect.attrib['y'] = fmt(ymin)
            rect.attrib['width'] = fmt(width)
            rect.attrib['height'] = fmt(height)
            rect.attrib['fill'] = 'none'
            rect.attrib['stroke'] = 'red'
            circ = ET.SubElement(word, 'circle')
            circ.attrib['cx'] = fmt(-xofst)
            circ.attrib['cy'] = fmt(-yofst)
            circ.attrib['r'] = '3'
            circ.attrib['fill'] = 'red'
            circ.attrib['stroke'] = 'red'
        bbox = (xy[0], xy[0]+width, xy[1]+ymin, xy[1]+ymax)

        if self.rotation:
            centerx = xy[0]  # Center of rotation
            centery = xy[1]
            costh = math.cos(math.radians(self.rotation))
            sinth = math.sin(math.radians(self.rotation))
            p1 = (xofst, ymin+yofst)  # Corners relative to rotation point
            p2 = (xofst+width, ymin+yofst)
            p3 = (xofst+width, ymax+yofst)
            p4 = (xofst, ymax+yofst)
            x1 = centerx + (p1[0]*costh + p1[1]*sinth) - xofst
            x2 = centerx + (p2[0]*costh + p2[1]*sinth) - xofst
            x3 = centerx + (p3[0]*costh + p3[1]*sinth) - xofst
            x4 = centerx + (p4[0]*costh + p4[1]*sinth) - xofst
            y1 = centery - (p1[0]*sinth - p1[1]*costh) - yofst
            y2 = centery - (p2[0]*sinth - p2[1]*costh) - yofst
            y3 = centery - (p3[0]*sinth - p3[1]*costh) - yofst
            y4 = centery - (p4[0]*sinth - p4[1]*costh) - yofst
            bbox = (min(x1, x2, x3, x4), max(x1, x2, x3, x4),
                    min(y1, y2, y3, y4), max(y1, y2, y3, y4))

            if self.rotation_mode == 'default':
                dx = {'left': x - bbox[0],
                      'right': x - bbox[1],
                      'center': x - (bbox[1]+bbox[0])/2}.get(self.halign, 0)
                dy = {'top': y - bbox[2],
                      'bottom': y - bbox[3],
                      'base': -sinth*dx,
                      'center': y - (bbox[3]+bbox[2])/2}.get(self.valign, 0)
                xform = f'translate({xy[0]+dx} {xy[1]+dy})'
                bbox = (bbox[0]+dx, bbox[1]+dx,
                        bbox[2]+dy, bbox[3]+dy)
            xform += f' rotate({-self.rotation} {-xofst} {-yofst})'

        if config.debug:
            rect = ET.SubElement(svg, 'rect')
            rect.attrib['x'] = fmt(bbox[0])
            rect.attrib['y'] = fmt(bbox[2])
            rect.attrib['width'] = fmt(bbox[1]-bbox[0])
            rect.attrib['height'] = fmt(bbox[3]-bbox[2])
            rect.attrib['fill'] = 'none'
            rect.attrib['stroke'] = 'blue'

        word.attrib['transform'] = xform
        if self.color:
            word.attrib['fill'] = self.color
        return svg, bbox

    def _buildstring(self) -> Symbols:
        ''' Create symbols and svg word in a <g> group tag, for placing in an svg '''
        scale = self.size / self.font.info.layout.unitsperem
        lineheight = self.size * self.linespacing
        lines = self.str.splitlines()
        yvals = [i*lineheight for i in range(len(lines))]  # valign == 'base'

        # Generate symbols and calculate x positions using left alignment
        symbols: list[ET.Element] = []  # <symbol> elements
        linewidths: list[float] = []
        allglyphs = []  # (glyph, x) where x is left aligned
        for lineidx, line in enumerate(lines):
            lineglyphs = []
            glyphs = [self.font.glyph(c) for c in line]
            x = 0
            for gidx, glyph in enumerate(glyphs):
                if glyph.id not in [s.attrib['id'] for s in symbols]:
                    symbols.append(glyph.svgsymbol())
                lineglyphs.append((glyph, x))
                nextglyph = glyphs[gidx+1] if gidx+1 < len(glyphs) else None
                xadvance = glyph.advance(nextglyph, kern=self.kern)
                x += (xadvance - min(0, glyph.path.bbox.xmin)) * scale

            if glyph.path.bbox.xmax > xadvance:
                # Make a bit wider to grab right edge that extends beyond advance width
                x += (glyph.path.bbox.xmax - xadvance) * scale
            linewidths.append(x)
            allglyphs.append(lineglyphs)

        # Place the glyphs based on halign
        word = ET.Element('g')
        word.attrib['word'] = self.str  # Just an identifier for debugging
        totwidth = max(linewidths)
        for lineidx, (lineglyphs, linewidth) in enumerate(zip(allglyphs, linewidths)):
            if self.halign == 'center':
                leftshift = (totwidth - linewidth)/2
            elif self.halign == 'right':
                leftshift = totwidth - linewidth
            else:  # halign = 'left'
                leftshift = 0
            for glyph, x in lineglyphs:
                word.append(glyph.place(x+leftshift, yvals[lineidx], self.size))

        ymin = yvals[0] - self.font.info.layout.ascent*scale
        ymax = yvals[-1] - self.font.info.layout.descent*scale

        if not config.svg2:
            symbols = []
        return Symbols(word, symbols, totwidth, ymin, ymax)

    def getsize(self) -> tuple[float, float]:
        ''' Calculate width and height (including ascent/descent) of string '''
        return self._symbols.width, self._symbols.ymax-self._symbols.ymin

    def getyofst(self) -> float:
        ''' Y-shift from bottom of bbox to 0 '''
        return -self._symbols.ymax


def _build_fontlist():
    ''' Generate list of system fonts locations and their names '''
    if sys.platform.startswith('win'):
        paths = [Path(r'C:\Windows\Fonts'),
                 Path(os.path.expandvars(r'%APPDATA%\Microsoft\Windows\Fonts'))]
    elif sys.platform.startswith('darwin'):
        paths = [Path().home() / 'Library/Fonts',
                 Path('/Library/Fonts/'),
                 Path('/System/Library/Fonts')]
    else:
        paths = [Path('/usr/share/fonts/'),
                 Path('/usr/local/share/fonts/'),
                 Path().home() / '.fonts',
                 Path().home() / '.local/share/fonts']

    fontlist = []
    for p in paths:
        fontlist.extend(p.rglob('*.ttf'))
        fontlist.extend(p.rglob('*.otf'))

    findex = {}
    for fname in fontlist:
        try:
            f = Font(fname)
        except ValueError:
            continue  # Unsupported Font
        family = f.info.names.family.lower()
        subfamily = f.info.names.subfamily.lower()
        subfamily = subfamily.replace('book', 'regular').replace('normal', 'regular')
        subfamily = subfamily.replace('italique', 'italic').replace('gras', 'bold')
        findex[(family, subfamily)] = fname
    return findex


def findfont(name, style='Regular'):
    ''' Find a font file by name '''
    if name is None:
        return None

    if Path(name).exists():
        return Path(name)

    global fontindex
    if fontindex is None:
        fontindex = _build_fontlist()
    return fontindex.get((name.lower(), style.lower()), None)



''' Glyph classes '''

from __future__ import annotations
from typing import Union

import os
import xml.etree.ElementTree as ET

from .fonttypes import GlyphPath, GlyphComp, BBox, Xform
from .config import config


def fmt(f: float) -> str:
    ''' String format, stripping trailing zeros '''
    p = f'.{config.precision}f'
    s = format(float(f), p)
    return s.rstrip('0').rstrip('.')  # Strip trailing zeros


def read_glyph(glyphid: int, font):
    ''' Read a glyph from the glyf table. '''
    offset = font._glyphoffset(glyphid)

    if offset is None:
        return EmptyGlyph(glyphid, font)

    if offset >= font.tables['glyf'].offset + font.tables['glyf'].length:
        return EmptyGlyph(glyphid, font)

    assert offset >= font.tables['glyf'].offset
    assert offset < font.tables['glyf'].offset + font.tables['glyf'].length

    font.fontfile.seek(offset)

    numcontours = font.fontfile.readint16()
    xmin = font.fontfile.readint16()
    ymin = font.fontfile.readint16()
    xmax = font.fontfile.readint16()
    ymax = font.fontfile.readint16()
    charbox = BBox(xmin, xmax, ymin, ymax)
    glyph: Union[SimpleGlyph, CompoundGlyph]
    if numcontours == -1:
        glyph = read_compoundglyph(font, glyphid, charbox)
    else:
        glyph = read_simpleglyph(font, glyphid, numcontours, charbox)
    return glyph


def read_simpleglyph(font, index, numcontours, charbox):
    ''' Read a symple glyph from the fontfile. Assumes file pointer is set '''
    fontfile = font.fontfile
    ends = []
    for i in range(numcontours):
        ends.append(font.fontfile.readuint16())
    instlength = fontfile.readuint16()
    fontfile.seek(instlength + fontfile.tell())  # Skip instructions

    numpoints = max(ends) + 1

    # flags
    ONCURVE = 0x01
    XSHORT = 0x02
    YSHORT = 0x04
    REPEAT = 0x08
    XDUP = 0x10
    YDUP = 0x20

    flags = []
    i = 0
    while i < numpoints:
        flag = fontfile.readuint8()
        if (flag & REPEAT):
            nrepeats = fontfile.readuint8() + 1  # Include the first one too
        else:
            nrepeats = 1
        flags.extend([flag] * nrepeats)
        i += nrepeats

    ctvals = [(f & ONCURVE) == 0 for f in flags]  # True for control point, False for real point
    xvals = []
    xval = 0  # Points are stored as deltas. Add them up as we go to get real points.
    for flag in flags:
        if (flag & XSHORT):  # X is one-byte
            if (flag & XDUP):
                xval += fontfile.readuint8()
            else:
                xval -= fontfile.readuint8()
        elif not (flag & XDUP):
            xval += fontfile.readint16()
        # else: xval stays the same
        xvals.append(xval)

    yvals = []
    yval = 0  # Add up deltas
    for flag in flags:
        if (flag & YSHORT):  # Y is one-byte
            if (flag & YDUP):
                yval += fontfile.readuint8()
            else:
                yval -= fontfile.readuint8()
        elif not (flag & YDUP):
            yval += fontfile.readint16()
        # else: yval stays the same
        yvals.append(yval)

    path = GlyphPath(xvals, yvals, ctvals, ends, charbox)
    glyph = SimpleGlyph(index, path, font)
    return glyph


def read_compoundglyph(font, index, charbox):
    ''' Read compound glyph from the fontfile. Assumes filepointer is set '''
    fontfile = font.fontfile

    ARG_WORDS = 0x0001
    ARG_XY = 0x0002
    # ROUND = 0x0004
    SCALE = 0x0008
    MORE = 0x0020
    XYSCALE = 0x0040
    TWOBYTWO = 0x0080
    # INSTRUCTIONS = 0x0100
    # METRICS = 0x0200
    # OVERLAP = 0x0400

    glyphidxs = []
    xforms = []
    moreglyphs = True
    while moreglyphs:
        flag = fontfile.readuint16()
        moreglyphs = (flag & MORE) == MORE
        subindex = fontfile.readuint16()
        match = False
        if (flag & ARG_WORDS):
            if (flag & ARG_XY):
                e = fontfile.readint16()
                f = fontfile.readint16()
            else:
                match = True
                e = fontfile.readuint16()
                f = fontfile.readuint16()
        else:
            if (flag & ARG_XY):
                e = fontfile.readint8()
                f = fontfile.readint8()
            else:
                match = True
                e = fontfile.readint8()
                f = fontfile.readint8()

        # Read Transformation
        if (flag & SCALE):
            a = d = fontfile.readshort()
            b = c = 0.
        elif (flag & XYSCALE):
            a = fontfile.readshort()
            d = fontfile.readshort()
            b = c = 0.
        elif (flag & TWOBYTWO):
            a = fontfile.readshort()
            b = fontfile.readshort()
            c = fontfile.readshort()
            d = fontfile.readshort()
        else:
            a = d = 1.
            b = c = 0.

        xforms.append(Xform(a, b, c, d, e, f, match))
        glyphidxs.append(subindex)

    glyphs = []
    for idx in glyphidxs:
        glyphs.append(read_glyph(idx, font))

    comp = GlyphComp(glyphs, xforms, charbox)
    return CompoundGlyph(index, comp, font)


class SimpleGlyph:
    ''' Simple Glyph '''
    dfltsize = 12   # Draw <symbols> in this point size

    def __init__(self, index: int, path: GlyphPath, font, char: str=None):
        self.char = char
        self.index = index
        self.path = path
        self.font = font
        basename, _ = os.path.splitext(os.path.basename(self.font.info.filename))
        self.id = f'{basename}_{index}'
        self.emscale = self.dfltsize / self.font.info.layout.unitsperem

    def _repr_svg_(self):
        return ET.tostring(self.svgxml(), encoding='unicode')

    def place(self, x, y, fontsize):
        ''' Get <use> svg tag translated/scaled to the right position '''
        fntscale = (fontsize/self.dfltsize)
        yshift = self.font.info.layout.ymax * self.emscale * fntscale
        if config.svg2:
            elm = ET.Element('use')
            elm.attrib['href'] = f'#{self.id}'
            elm.attrib['transform'] = f'translate({fmt(x)} {fmt(y-yshift)}) scale({fmt(fntscale)})'
        else:
            elm = self.svgpath(x0=x, y0=y, scale=fntscale)
        return elm

    def advance(self, nextchr=None, kern=True):
        ''' Get advance width in glyph units, including kerning if nextchr is defined '''
        if nextchr:
            nextchr = nextchr.index
        return self.font.advance(self.index, nextchr, kern=kern)

    def svgpath(self, x0=0, y0=0, scale=1) -> ET.Element:
        ''' Get svg <path> element for glyph, normalized to 12-point font '''
        emscale = self.emscale * scale
        # Split the contours
        xconts = []
        yconts = []
        ctrls = []
        start = 0
        for i in range(len(self.path.ends)):
            stop = self.path.ends[i]+1
            xconts.append(self.path.xvals[start:stop])
            yconts.append(self.path.yvals[start:stop])
            ctrls.append(self.path.ctvals[start:stop])
            start = stop

        path = ''
        for xvals, yvals, ctrl in zip(xconts, yconts, ctrls):
            xx = [x0 + x*emscale for x in xvals]
            yy = [y0 - y*emscale for y in yvals]
            npoints = len(xx)

            if ctrl[0]:
                # Path STARTS with a control point. Wrap last point in path.
                # Unna-Regular.ttf is an example.
                if ctrl[1]:
                    xim = (xx[0] + xx[1])/2
                    yim = (yy[0] + yy[1])/2
                    path += f'M {fmt(xx[-1])} {fmt(yy[-1])} Q {fmt(xx[0])} {fmt(yy[0])} {fmt(xim)} {fmt(yim)}'
                else:
                    path += f'M {fmt(xx[-1])} {fmt(yy[-1])} Q {fmt(xx[0])} {fmt(yy[0])} {fmt(xx[1])} {fmt(yy[1])}'
            else:
                path += f'M {fmt(xx[0])} {fmt(yy[0])} '

            i = 1
            while i < npoints:
                if ctrl[i]:
                    if i == npoints-1:
                        # Last point is control. End point wraps to start point
                        path += f'Q {fmt(xx[i])} {fmt(yy[i])}, {fmt(xx[0])} {fmt(yy[0])} '
                        i += 1
                    elif ctrl[i+1]:
                        # Next point is also control.
                        # End of this bezier is implied between two controls
                        xim = (xx[i] + xx[i+1])/2
                        yim = (yy[i] + yy[i+1])/2
                        path += f'Q {fmt(xx[i])} {fmt(yy[i])}, {fmt(xim)} {fmt(yim)} '
                        i += 1
                    else:
                        # Next point is real. It's the endpoint.
                        path += f'Q {fmt(xx[i])} {fmt(yy[i])}, {fmt(xx[i+1])} {fmt(yy[i+1])} '
                        i += 2
                else:
                    path += f'L {fmt(xx[i])} {fmt(yy[i])} '
                    i += 1

            path += 'Z '
        return ET.Element('path', attrib={'d': path})

    def svgsymbol(self) -> ET.Element:
        ''' Get svg <symbol> element for this glyph, scaled to 12-point font '''
        xmin = min(self.path.bbox.xmin * self.emscale, 0)
        xmax = self.path.bbox.xmax
        width = xmax-xmin
        ymax = max(self.font.info.layout.ymax, self.path.bbox.ymax) * self.emscale
        ymin = min(self.font.info.layout.ymin, self.path.bbox.ymin) * self.emscale
        height = ymax - ymin

        sym = ET.Element('symbol')
        sym.attrib['id'] = self.id
        sym.attrib['width'] = fmt(width)
        sym.attrib['height'] = fmt(height)
        sym.attrib['viewBox'] = f'{fmt(xmin)} {fmt(-ymax)} {fmt(width)} {fmt(height)}'
        sym.append(self.svgpath())
        return sym

    def svg(self, fontsize: float=None, svgver=2) -> str:
        ''' Get SVG as string '''
        return ET.tostring(self.svgxml(fontsize, svgver=svgver),
                           encoding='unicode')

    def svgxml(self, fontsize: float=None, svgver=2) -> ET.Element:
        ''' Standalong SVG '''
        fontsize = fontsize if fontsize else config.fontsize
        scale = fontsize / self.font.info.layout.unitsperem

        # Width varies by character, but height is constant for the whole font
        # View should include whole character, even if it goes negative/outside the advwidth
        xmin = min(self.path.bbox.xmin * scale, 0)
        xmax = self.path.bbox.xmax * scale
        ymin = min(self.path.bbox.ymin, self.font.info.layout.ymin) * scale

        # ymax can go above font's ymax for extended (ie math) glyphs
        ymax = max(self.path.bbox.ymax, self.font.info.layout.ymax) * scale
        width = xmax - xmin
        height = ymax - ymin
        base = ymax
        scale = fontsize/self.dfltsize

        svg = ET.Element('svg')
        svg.attrib['width'] = fmt(width)
        svg.attrib['height'] = fmt(height)
        svg.attrib['xmlns'] = 'http://www.w3.org/2000/svg'
        if not config.svg2:
            svg.attrib['xmlns:xlink'] = 'http://www.w3.org/1999/xlink'
            elm = self.svgpath(x0=xmin, y0=base, scale=scale)
            svg.append(elm)
        else:
            svg.attrib['viewBox'] = f'{fmt(xmin)} 0 {fmt(width)} {fmt(height)}'
            symbol = self.svgsymbol()
            svg.append(symbol)
            g = ET.SubElement(svg, 'use')
            g.attrib['href'] = f'#{self.id}'
            g.attrib['transform'] = f'translate({fmt(xmin)}, {fmt(base-ymax)}) scale({fmt(scale)})'
        return svg

    def test(self) -> 'TestGlyph':
        ''' Get Glyph Test representation showing vertices and borders '''
        return TestGlyph(self)


class CompoundGlyph(SimpleGlyph):
    ''' Compound glyph, made of multiple other Glyphs '''
    def __init__(self, index: int, glyphs: GlyphComp, font, char: str=None):
        self.char = char
        self.index = index
        self.glyphs = glyphs
        path = self._buildcontour()
        super().__init__(index, path, font, char)

    def _buildcontour(self) -> GlyphPath:
        ''' Combine multiple glyphs into one set of contours '''
        xvals: list[int] = []
        yvals: list[int] = []
        ctvals: list[int] = []
        ends: list[int] = []
        for glyph, xform in zip(self.glyphs.glyphs, self.glyphs.xforms):
            if xform.match:
                raise NotImplementedError('Compound glyph match transform')

            m0 = max(abs(xform.a), abs(xform.b))
            n0 = max(abs(xform.c), abs(xform.d))
            m = 2*m0 if abs(abs(xform.a)-abs(xform.c)) <= 33/65536 else m0
            n = 2*n0 if abs(abs(xform.b)-abs(xform.d)) <= 33/65536 else n0
            gx = [m * (xform.a/m * xx + xform.c/m * yy + xform.e) for xx, yy in zip(glyph.path.xvals, glyph.path.yvals)]
            gy = [n * (xform.b/n * xx + xform.d/n * yy + xform.f) for xx, yy in zip(glyph.path.xvals, glyph.path.yvals)]
            ends.extend([end + len(xvals) for end in glyph.path.ends])
            xvals.extend(gx)
            yvals.extend(gy)
            ctvals.extend(glyph.path.ctvals)
        path = GlyphPath(xvals, yvals, ctvals, ends, self.glyphs.bbox)
        return path


class TestGlyph:
    ''' Draw glyph svg with test/debug lines '''
    def __init__(self, glyph: SimpleGlyph):
        self.glyph = glyph

    def _repr_svg_(self):
        ''' Jupyter representation '''
        return self.svg()

    def svg(self, fontsize: float=None) -> str:
        ''' Glyph SVG string '''
        return ET.tostring(self.svgxml(fontsize), encoding='unicode')

    def svgxml(self, fontsize: float=None) -> ET.Element:
        ''' Glyph svg as XML element tree '''
        fontsize = fontsize if fontsize else config.fontsize
        svg = self.glyph.svgxml(fontsize)
        scale = fontsize / self.glyph.font.info.layout.unitsperem
        xmin = min(self.glyph.path.bbox.xmin * scale, 0)
        xmax = self.glyph.path.bbox.xmax * scale
        ymin = min(self.glyph.path.bbox.ymin, self.glyph.font.info.layout.ymin) * scale
        ymax = max(self.glyph.path.bbox.ymax, self.glyph.font.info.layout.ymax) * scale
        width = xmax - xmin
        height = ymax - ymin
        base = ymax  # = height - ymin

        # Borders and baselines
        path = ET.SubElement(svg, 'path')
        path.attrib['d'] = f'M {fmt(xmin)} {fmt(base)} L {fmt(width)} {fmt(base)}'
        path.attrib['stroke'] = 'red'

        ascent = base - self.glyph.font.info.layout.ascent * scale
        descent = base - self.glyph.font.info.layout.descent * scale
        path = ET.SubElement(svg, 'path')
        path.attrib['d'] = f'M {fmt(xmin)} {fmt(ascent)} L {fmt(width)} {fmt(ascent)}'
        path.attrib['stroke'] = 'gray'
        path.attrib['stroke-dasharray'] = '2 2'
        path = ET.SubElement(svg, 'path')
        path.attrib['d'] = f'M {fmt(xmin)} {fmt(descent)} L {fmt(width)} {fmt(descent)}'
        path.attrib['stroke'] = 'gray'
        path.attrib['stroke-dasharray'] = '2 2'
        rect = ET.SubElement(svg, 'rect')
        rect.attrib['x'] = '0'
        rect.attrib['y'] = '0'
        rect.attrib['width'] = fmt(xmax)
        rect.attrib['height'] = fmt(height)
        rect.attrib['fill'] = 'none'
        rect.attrib['stroke'] = 'blue'
        rect.attrib['stroke-dasharray'] = '2 2'
        circ = ET.SubElement(svg, 'circle')
        circ.attrib['cx'] = '0'
        circ.attrib['cy'] = fmt(base)
        circ.attrib['r'] = '3'
        circ.attrib['fill'] = 'red'

        # Dots defining <path>
        for x, y, c in zip(self.glyph.path.xvals, self.glyph.path.yvals, self.glyph.path.ctvals):
            circ = ET.SubElement(svg, 'circle')
            circ.attrib['cx'] = fmt(x * scale)
            circ.attrib['cy'] = fmt(base - y * scale)
            circ.attrib['r'] = f'{fmt(fontsize*scale/3)}'
            circ.attrib['fill'] = 'none' if c else 'blue'
            circ.attrib['stroke'] = 'blue'
            circ.attrib['opacity'] = '0.3'
        return svg


class EmptyGlyph(SimpleGlyph):
    ''' Glyph with no contours (like a space) '''
    def __init__(self, index: int, font):
        path = GlyphPath([], [], [], [], BBox(0, 0, 0, 0))
        super().__init__(index, path, font)



''' Global configuration options '''


class Config:
    ''' Global configuration options for Ziafont

        Attributes
        ----------
        svg2: Use SVG2.0. Disable for better browser compatibility
            at the expense of SVG size
        debug: Debug mode, draws bounding boxes around text
        fontsize: Default font size in points
        precision: Decimal precision for SVG coordinates
    '''
    svg2: bool = True
    debug: bool = False
    fontsize: float = 48
    precision: float = 3


config = Config()



from collections import namedtuple

Table = namedtuple('Table', ['checksum', 'offset', 'length'])
BBox = namedtuple('BBox', ['xmin', 'xmax', 'ymin', 'ymax'])
AdvanceWidth = namedtuple('AdvanceWidth', ['width', 'leftsidebearing'])
Header = namedtuple(
    'Header', ['version', 'revision', 'checksumadjust', 'magic', 'flags',
               'created', 'modified', 'macstyle', 'lowestrecppem',
               'directionhint', 'indextolocformat', 'glyphdataformat', 'numlonghormetrics'])
FontNames = namedtuple('FontNames', ['copyright', 'family', 'subfamily', 'unique', 'name', 'nameversion',
                       'postscript', 'trademark', 'manufacturer', 'designer',
                       'description', 'vendorurl', 'designerurl', 'license', 'LicenseURL'])
Layout = namedtuple(
    'Layout', ['unitsperem', 'xmin', 'xmax', 'ymin', 'ymax',
               'ascent', 'descent', 'advwidthmax',
               'minleftbearing', 'minrightbearing'])
FontInfo = namedtuple(
    'FontInfo', ['filename', 'names', 'header', 'layout'])

GlyphPath = namedtuple('GlyphPath', ['xvals', 'yvals', 'ctvals', 'ends', 'bbox'])
GlyphComp = namedtuple('GlyphComp', ['glyphs', 'xforms', 'bbox'])
Xform = namedtuple('Xform', ['a', 'b', 'c', 'd', 'e', 'f', 'match'])
Symbols = namedtuple('Symbols', ['word', 'symbols', 'width', 'ymin', 'ymax'])



''' Font CMAP table - translate character into a glyph ID '''

from typing import Sequence


class Cmap12:
    ''' Format 12 cmap Table '''
    def __init__(self, platform: str, platformid: int, startcodes: Sequence[int],
                 endcodes: Sequence[int], glyphstarts: Sequence[int]):
        self.platform = platform
        self.platformid = platformid
        self.startcodes = startcodes
        self.endcodes = endcodes
        self.glyphstarts = glyphstarts

    def __repr__(self):
        return f'<Cmap Format 12: {self.platform} id={self.platformid}'

    def glyphid(self, char: str) -> int:
        ''' Get glyph index from a character '''
        charid = ord(char)

        for endid, end in enumerate(self.endcodes):
            if end >= charid:
                break
        else:
            return 0  # Character not found

        if self.startcodes[endid] > charid:
            return 0

        # Char was found
        ofst = charid - self.startcodes[endid]
        return self.glyphstarts[endid] + ofst


class Cmap4:
    ''' Format 4 cmap Table '''
    def __init__(self, platform: str, platformid: int, startcodes: Sequence[int],
                 endcodes: Sequence[int], idrangeoffset: Sequence[int],
                 iddeltas: Sequence[int], glyphidarray: Sequence[int]):
        self.platform = platform
        self.platformid = platformid
        self.startcodes = startcodes
        self.endcodes = endcodes
        self.idrangeoffset = idrangeoffset
        self.iddeltas = iddeltas
        self.glyphidarray = glyphidarray

    def __repr__(self):
        return f'<Cmap Format 4: {self.platform} id={self.platformid}'

    def glyphid(self, char: str) -> int:
        ''' Get glyph index from a character '''
        charid = ord(char)

        for endid, end in enumerate(self.endcodes):
            if end >= charid:
                break
        else:
            return 0  # Character not found (missing glyph)

        if self.startcodes[endid] > charid:
            return 0

        if self.idrangeoffset[endid] != 0:
            segcount = len(self.startcodes)
            idx = endid - segcount + self.idrangeoffset[endid]//2 + (charid - self.startcodes[endid])
            gid = self.glyphidarray[idx]
            glyphid = self.iddeltas[endid] + gid if gid > 0 else 0
        else:
            glyphid = (self.iddeltas[endid] + charid) % 0x10000
        return glyphid
